/*
    Copyright (C) 2002, Steve Lhomme <steve.lhomme @ free.fr>
    This file is part of ac32wav.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/**
    \file processus.cpp
    Convert AC3 files to RIFF/WAV/AC3 files
    \todo remove bad frames
    \todo keep bad data
    \note see http://www.atsc.org/standards/a_52a.pdf
    \note see http://www.microsoft.com/hwdev/tech/audio/multichaud.asp
    \note see http://www.intelligraphics.com/articles/WDMaudiodesign_article.html
*/
#include <stdio.h>
#include <cassert>

#include "processus.h"
#include "ac3file.h"
#include "ac3iframe.h"
#include "ac3frame.h"
#include "ac3_wav.h"

#if !defined(NO_NAMESPACE)
using namespace std;
using namespace M2W_NAMESPACE;
#endif

const uint32 BUFFER_SIZE = 5000;

#ifdef  __cplusplus
extern "C" {
#endif

int process(const string & input_Filename, const string & output_Filename, bool bKeepAll = true, bool bVerbose = true, bool bVerboseMore = true)
{
    if (bVerbose)
	printf("\r\nConverting %s to %s\r\n", input_Filename.c_str(), output_Filename.c_str());

    // open the input file
    AC3_File input(input_Filename, mode_read);

    // open the output file
    AC3_WAV_file output(output_Filename);
    if (!output.Open(FILE_MODE_WRITE | FILE_MODE_ERASE))
    {
	printf("Impossible to open output file '%s'\r\n", output_Filename.c_str());
	return -1;
    }

    // prepare to read the whole file for statistics
    if (bVerbose)
	printf("Reading statistics in %s\r\n", input_Filename.c_str());

    input.SkipToFrameStart();

    bool bInfoFrameOn = false;
/*    MPA_Lame_Frame infoFrame;
    if (input.CheckInfoFrame())
    {
	*static_cast<AC3_Frame *>(&infoFrame) = input.GetInfoFrame();
	bInfoFrameOn = true;
    }
*/
    uint32 FrameNumber = 0;
    uint64 SampleNumbers = 0;
    uint64 TotalFrameSize = 0;
    AC3_Frame frame;

    while (input.FindNextSyncWord())
    {
	frame = input.GetFrame();
	FrameNumber++;
	if (input.Feed())
	{
	    /// \todo handle the bitrate of the frame in a key/value table, so that we can detect real CBR files
	    TotalFrameSize += frame.DataSize();
	    SampleNumbers += frame.SampleSize();
	    uint16 frame_size = frame.DataSize();
	    uint16 skipped = input.SkipData();
	    if (frame_size != skipped)
	    {
		TotalFrameSize -= frame_size;
		TotalFrameSize += skipped;
	    }
	}
	else input.SkipData();
    }
    input.FixStats();

    if (bVerboseMore)
    {
	printf("Sampling Rate = %d / Total Valid Frame Size = %ld\r\n",
	       input.SamplingRate(), TotalFrameSize);
	printf("Average Bitrate = %d (8 x %d)\r\n", input.AverageBitRate(), uint32(0.5 + (input.AverageBitRate() / 8)));
	printf("Sample number = %d\r\n", SampleNumbers);
	printf("FrameNumber = %d\r\n", FrameNumber);
    }

    // construct the WAV header with the statistics

    // prepare to read for output
    input.SkipToFrameStart();

    /// \todo add the option to skip the Info frame or not
    if (bInfoFrameOn)
    {
	input.FindNextSyncWord();
	input.GetFrame();
	input.SkipData();
    }

    if (bVerbose)
	printf("Writing frames in %s\r\n", output_Filename.c_str());

    /// \todo move this somewhere else (cleaner)
    output.setFrameType(input.FrameType());
    if (input.IsCBR())
	output.setAlignment(input.FrameType().DataSize());
    else
	output.setAlignment(1);
    output.setByterate(0.5 + float(input.AverageBitRate() >> 3));
//    output.setSampleSize(SampleSize); // should be computed wether it's CBR or not

    output.WriteHead(TotalFrameSize, SampleNumbers, false);

    binary frame_buffer[BUFFER_SIZE]; // should be enough for 32kHz/320kbps
    if (!bKeepAll)
    {
	while (input.FindNextSyncWord())
	{
	    frame = input.GetFrame();
	    if (input.Test())
	    {
		output.Write(frame, 4);
		assert(frame.DataSize() < BUFFER_SIZE);
		input.GetData(frame_buffer, BUFFER_SIZE);
		output.Write(frame_buffer, frame.DataSize());
	    }
	    else input.SkipData();
	}
    }
    else
    {
	uint32 sizeRead;
	do
	{
	    sizeRead = input.read(frame_buffer, BUFFER_SIZE);
	    output.Write(frame_buffer, sizeRead);
	} while (sizeRead == BUFFER_SIZE);
    }

    output.Close();

    return 0;
}

#ifdef  __cplusplus
}
#endif
